﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace PMDMikuDirect3d
{
    public partial class SimpleLiteD3d : IDisposable
    {
       public struct PMD_Header     //【PMDファイル ヘッダー情報】
       {  public byte[] ID;         //ID 3バイト "P","m","d"
          public float vNum;        // バージョン番号
          public byte[] ModelName;  //モデル名称 20バイト
          public byte[] CopyRight;  //著作権情報 256バイト
          public string sID;        //●Shift_JIS
          public string sModelName; //●Shift_JIS
          public string sCopyRight; //●Shift_JIS
       }

       public struct PMD_vertex     //【PMD_頂点データ】　4*3+4*3+4*2+2*2+2 = 12+12+8+4+2 = 38bytes/頂点
       {  public int nPoints;       //【頂点個数】 
          public CustomVertex.PositionNormalTextured[] UserVertex; //【頂点座標】【法線ﾍﾞｸﾄﾙ】【uv座標】
          public CustomVertex.PositionNormalTextured[] OriginalVertex;  //【頂点座標】【法線ﾍﾞｸﾄﾙ】【uv座標】SphereMap対応
          public short[] Bone1Num;  //【Bone番号】Bone番号1 モデル変形(頂点移動)時に影響
          public short[] Bone2Num;  //【Bone番号】Bone番号2      :
          public Byte[] BoneWeight; //【Bone重み係数】Bone1に与える影響度 0-100 Bone2への影響度は(100-bone_weight)で与えられる。
          public Byte[] EdgeFlag;   //【ｴｯｼﾞ･ﾌﾗｸﾞ】0:通常、1:エッジ無効　輪郭線が有効の場合
       }

       public struct PMD_index        //【PMD_インデックスデータ(面情報)】MikuMikuDanceのモデルではtriangle(三角形)しか使っていません。
       {  public int nIndex;          //【配列要素数】＝面数(三角形の個数)*3
          public int nPolygons;       //●三角形の個数
          public short[] IndexBuffer; //【頂点番号リスト】３個/面
       } //注）本質的にはGPU側の制約によるが、この配列のためにMMDでは最大65535頂点までのモデルしか取り扱えない。

       public struct PMD_material          //【PMD_材質データ】
       {  public int nMaterial;            //【材質個数】 
          public Material[] Material;      //【材質データ】DirectXの材質ﾃﾞｰﾀ構造体と同じ
          public byte[] ToonIndex;         //【ﾄｩｰﾝ･ｲﾝﾃﾞｯｸｽ番号】toon??.bmp 0.bmp:0xFF, 1.bmp:0x00, ･･･10.bmp:0x09
          public byte[] EdgeFlag;          //【ｴｯｼﾞ･ﾌﾗｸﾞ】輪郭､影
          public int[] nPoints;            //【面構成要素数】面数*3 = 面頂点数 実際のindexに直すには材質0から累積加算する。
          public string[] TextureFileName; //【ﾃｸｽﾁｬ･ﾌｧｲﾙ名】20bytesギリギリまで使える 20bytes時は0x00がなくても可。'
          public Texture[] Texture;        //【ﾃｸｽﾁｬ】
          public bool[] SphereMap;         //【SphereMapフラグ】
       }

       public PMD_Header   PMDh; //【PMDヘッダー情報】
       public PMD_vertex   PMDv; //【PMD頂点バッファ】
       public PMD_index    PMDi; //【PMDインデックス】
       public PMD_material PMDm; //【PMD材質情報】
       public byte[] bShortName = new byte[20]; //【テクスチャーファイル名の一時バッファ】
       public const float Zoom = 20.0f; //【変換倍率】座標系をMMD⇒ARに変換する際、1ミクセルを20倍にして表示。
       public const string PMDPath = "../../PMD/"; //【PMDファイルの相対パス】

       //***********************************************************************************************
       //*** FileStreamとBinaryReaderを使ってPMDファイルからMMD構造体に情報を読み込む
       //***********************************************************************************************
       public void ReadPMDFile(string PMDFilePath) //【MikuMikuDanceのPMDファイルからHeader, Vertex, Index, Meterialを読み込む】
       {  
          FileStream PMDFile = new FileStream(PMDFilePath, FileMode.Open); //【PMDファイル読み込みの為のFileStream】    
          BinaryReader bRD = new BinaryReader(PMDFile);                    //【PMDファイル読み込みの為のBinaryReader】     

           //【PMDヘッダー情報を読み込む】
          PMDh.ID = new byte[3];          //ID 3バイト "P","m","d"
          PMDh.ModelName = new byte[20];  //モデル名称 20バイト
          PMDh.CopyRight = new byte[256]; //著作権情報 256バイト
          PMDh.ID = bRD.ReadBytes(3);          // ID 3バイト "P","m","d"を読み込む
          PMDh.vNum = bRD.ReadSingle();        // バージョン番号を読み込む
          PMDh.ModelName = bRD.ReadBytes(20);  // モデル名称を読み込む
          PMDh.CopyRight = bRD.ReadBytes(256); // 著作権情報を読み込む
          PMDh.sID = Encoding.GetEncoding(932).GetString(PMDh.ID);               // Shift JISに変換
          PMDh.sModelName = Encoding.GetEncoding(932).GetString(PMDh.ModelName); // Shift JISに変換
          PMDh.sCopyRight = Encoding.GetEncoding(932).GetString(PMDh.CopyRight); // Shift JISに変換

           //【PMD頂点情報を読み込む】
          PMDv.nPoints = bRD.ReadInt32(); //
          PMDv.UserVertex = new CustomVertex.PositionNormalTextured[PMDv.nPoints]; // Tu,TvはSphereMapで変更されることがある。
          PMDv.OriginalVertex = new CustomVertex.PositionNormalTextured[PMDv.nPoints];  //【Original】
           PMDv.Bone1Num = new short[PMDv.nPoints];  //【Bone番号】Bone番号1 モデル変形(頂点移動)時に影響
          PMDv.Bone2Num = new short[PMDv.nPoints];  //【Bone番号】Bone番号2      :
          PMDv.BoneWeight = new Byte[PMDv.nPoints]; //【Bone重み係数】Bone1に与える影響度 0-100 Bone2への影響度は(100-bone_weight)で与えられる。
          PMDv.EdgeFlag = new Byte[PMDv.nPoints];   //【ｴｯｼﾞ･ﾌﾗｸﾞ】0:通常、1:エッジ無効　輪郭線が有効の場合
          for (var i = 0; i < PMDv.nPoints; i++)    //【右手系から左手系に】Z⇒-Z, Nz⇒-Nz
          {  PMDv.UserVertex[i].Position = new Vector3(bRD.ReadSingle()*Zoom, bRD.ReadSingle()*Zoom, -bRD.ReadSingle()*Zoom); //【位置ベクトル】X,Y,Z
             PMDv.UserVertex[i].Normal   = new Vector3(bRD.ReadSingle(), bRD.ReadSingle(), -bRD.ReadSingle()); //【法線ベクトル】Nx, Ny, Nz
             PMDv.UserVertex[i].Tu       = bRD.ReadSingle(); //【テクスチャーUV空間】u
             PMDv.UserVertex[i].Tv       = bRD.ReadSingle(); //【テクスチャーUV空間】v
             PMDv.Bone1Num[i]   = bRD.ReadInt16();
             PMDv.Bone2Num[i]   = bRD.ReadInt16();
             PMDv.BoneWeight[i] = bRD.ReadByte();
             PMDv.EdgeFlag[i]   = bRD.ReadByte();
             PMDv.OriginalVertex[i] = PMDv.UserVertex[i]; //【ShpereMapでTu, Tvを壊すのでオリジナルを保存しておく】
          }

           //【PMD面(インデックス)情報を読み込む】
          PMDi.nIndex = bRD.ReadInt32();
          PMDi.nPolygons = PMDi.nIndex / 3;
          PMDi.IndexBuffer = new short[PMDi.nIndex];
          for (var i = 0; i < PMDi.nIndex; i++) { PMDi.IndexBuffer[i] = bRD.ReadInt16(); }

           //【PMD材質情報を読み込む】 
          PMDm.nMaterial = bRD.ReadInt32();
          PMDm.Material = new Material[PMDm.nMaterial];
          PMDm.ToonIndex = new byte[PMDm.nMaterial];
          PMDm.EdgeFlag = new byte[PMDm.nMaterial];
          PMDm.nPoints = new int[PMDm.nMaterial];
          PMDm.TextureFileName = new string[PMDm.nMaterial];
          PMDm.SphereMap = new bool[PMDm.nMaterial];
          for (var i = 0; i < PMDm.nMaterial; i++)
          {  PMDm.Material[i].AmbientColor = new ColorValue(bRD.ReadSingle(), bRD.ReadSingle(), bRD.ReadSingle(), bRD.ReadSingle()); //【環境色】　　Red, Green, Blue, Alpha
             PMDm.Material[i].SpecularSharpness = bRD.ReadSingle();                                                 //【鏡面反射強度】
             PMDm.Material[i].SpecularColor = new ColorValue(bRD.ReadSingle(), bRD.ReadSingle(), bRD.ReadSingle()); //【鏡面反射色】Red, Green, Blue 
             PMDm.Material[i].DiffuseColor  = new ColorValue(bRD.ReadSingle(), bRD.ReadSingle(), bRD.ReadSingle()); //【拡散光色】　Red, Green, Blue
             PMDm.Material[i].EmissiveColor = new ColorValue(0.0F, 0.0F, 0.0F); //【発光色はMikuMikuDanceでは未使用】
             PMDm.ToonIndex[i] = bRD.ReadByte();
             PMDm.EdgeFlag[i] = bRD.ReadByte();
             PMDm.nPoints[i] = bRD.ReadInt32();
             bShortName = bRD.ReadBytes(20);
             string  s = Encoding.GetEncoding(932).GetString(bShortName);
             PMDm.TextureFileName[i] = "";
             for (var k = 0; k < 20; k++)
             {  if (s[k] == 0) { break; } else { PMDm.TextureFileName[i] += s[k]; }
             }
             PMDm.SphereMap[i] = false;
             if (PMDm.TextureFileName[i] != "")
             {
                 string[] p = PMDm.TextureFileName[i].Split(new char[] { '.' });
                 if (p[1] == "sph") { PMDm.SphereMap[i] = true; } //【ファイル名がsph形式の場合】
                 PMDm.TextureFileName[i] = PMDPath + PMDm.TextureFileName[i]; 
             }
          }

          //【PMDファイルを閉じる】
          bRD.Close();
          PMDFile.Close();
       }    
    }
}
